from __future__ import division
from avro import schema, io
import abc
import six
import datetime
import decimal

long = int

EPOCH_DATE = datetime.date(1970, 1, 1)
SECONDS_IN_DAY = 24 * 60 * 60


class LogicalTypeProcessor(object, six.with_metaclass(abc.ABCMeta)):
    @abc.abstractmethod
    def validate(self, expected_schema, datum):
        pass

    @abc.abstractmethod
    def does_match(self, writers_schema, readers_schema):
        return False

    @abc.abstractmethod
    def convert_back(self, writers_schema, readers_schema, value):
        return value

    @abc.abstractmethod
    def convert(self, writers_schema, value):
        pass

    @abc.abstractmethod
    def can_convert(self, writers_schema):
        pass

    @abc.abstractmethod
    def typename(self):
        pass

    @abc.abstractmethod
    def initializer(self, value=None):
        pass


class DecimalLogicalTypeProcessor(LogicalTypeProcessor):
    def can_convert(self, writers_schema):
        return (
            isinstance(writers_schema, schema.PrimitiveSchema)
            and writers_schema.type == "string"
        )

    def validate(self, expected_schema, datum):
        return isinstance(datum, (int, float, long, decimal.Decimal))

    def convert(self, writers_schema, value):
        if not isinstance(value, (int, float, long, decimal.Decimal)):
            raise Exception("Wrong type for decimal conversion")
        return str(value)

    def convert_back(self, writers_schema, readers_schema, value):
        return decimal.Decimal(value)

    def does_match(self, writers_schema, readers_schema):
        if isinstance(writers_schema, schema.PrimitiveSchema):
            if writers_schema.type == "string":
                return True
        return False

    def typename(self):
        return "decimal.Decimal"

    def initializer(self, value=None):
        return "decimal.Decimal(%s)" % (0 if value is None else value)


class DateLogicalTypeProcessor(LogicalTypeProcessor):
    _matching_types = {"int", "long", "float", "double"}

    def can_convert(self, writers_schema):
        return (
            isinstance(writers_schema, schema.PrimitiveSchema)
            and writers_schema.type == "int"
        )

    def validate(self, expected_schema, datum):
        return isinstance(datum, datetime.date)

    def convert(self, writers_schema, value):
        if not isinstance(value, datetime.date):
            raise Exception("Wrong type for date conversion")
        return (value - EPOCH_DATE).total_seconds() // SECONDS_IN_DAY

    def convert_back(self, writers_schema, readers_schema, value):
        return EPOCH_DATE + datetime.timedelta(days=int(value))

    def does_match(self, writers_schema, readers_schema):
        if isinstance(writers_schema, schema.PrimitiveSchema):
            if writers_schema.type in DateLogicalTypeProcessor._matching_types:
                return True
        return False

    def typename(self):
        return "datetime.date"

    def initializer(self, value=None):
        return (
            ("logical.DateLogicalTypeProcessor().convert_back(None, None, %s)" % value)
            if value is not None
            else "datetime.datetime.today().date()"
        )


class TimeMicrosLogicalTypeProcessor(LogicalTypeProcessor):
    _matching_types = {"int", "long", "float", "double"}

    def can_convert(self, writers_schema):
        return (
            isinstance(writers_schema, schema.PrimitiveSchema)
            and writers_schema.type == "long"
        )

    def validate(self, expected_schema, datum):
        return isinstance(datum, datetime.time)

    def convert(self, writers_schema, value):
        if not isinstance(value, datetime.time):
            raise Exception("Wrong type for time conversion")
        return (
            (value.hour * 60 + value.minute) * 60 + value.second
        ) * 1000000 + value.microsecond

    def convert_back(self, writers_schema, readers_schema, value):
        (
            _,
            hours,
            minutes,
            seconds,
            microseconds,
        ) = TimeMicrosLogicalTypeProcessor.extract_time_parts(value)
        return datetime.time(hours, minutes, seconds, microseconds)

    @staticmethod
    def extract_time_parts(value):
        value = long(value)
        days = value // 86400000000
        hours = value % 86400000000
        minutes = hours % 3600000000
        hours = hours // 3600000000
        seconds = minutes % 60000000
        minutes = minutes // 60000000
        microseconds = seconds % 1000000
        seconds = seconds // 1000000
        return days, hours, minutes, seconds, microseconds

    def does_match(self, writers_schema, readers_schema):
        if isinstance(writers_schema, schema.PrimitiveSchema):
            if writers_schema.type in TimeMicrosLogicalTypeProcessor._matching_types:
                return True
        return False

    def typename(self):
        return "datetime.time"

    def initializer(self, value=None):
        return (
            (
                "logical.TimeMicrosLogicalTypeProcessor().convert_back(None, None, %s)"
                % value
            )
            if value is not None
            else "datetime.datetime.today().time()"
        )


class TimeMillisLogicalTypeProcessor(TimeMicrosLogicalTypeProcessor):
    def can_convert(self, writers_schema):
        return (
            isinstance(writers_schema, schema.PrimitiveSchema)
            and writers_schema.type == "int"
        )

    def convert(self, writers_schema, value):
        if not isinstance(value, datetime.time):
            raise Exception("Wrong type for time conversion")
        return int(
            super(TimeMillisLogicalTypeProcessor, self).convert(writers_schema, value)
            // 1000
        )

    def convert_back(self, writers_schema, readers_schema, value):
        return super(TimeMillisLogicalTypeProcessor, self).convert_back(
            writers_schema, readers_schema, value * 1000
        )

    def initializer(self, value=None):
        return (
            (
                "logical.TimeMillisLogicalTypeProcessor().convert_back(None, None, %s)"
                % value
            )
            if value is not None
            else "datetime.datetime.today().time()"
        )


class TimestampMicrosLogicalTypeProcessor(LogicalTypeProcessor):
    _matching_types = {"int", "long", "float", "double"}

    def can_convert(self, writers_schema):
        return (
            isinstance(writers_schema, schema.PrimitiveSchema)
            and writers_schema.type == "long"
        )

    def validate(self, expected_schema, datum):
        return isinstance(datum, datetime.datetime)

    def convert(self, writers_schema, value):
        if not isinstance(value, datetime.datetime):
            if isinstance(value, datetime.date):
                value = datetime.datetime(
                    value.year,
                    value.month,
                    value.day,
                    0,
                    0,
                    0,
                    0,
                    datetime.timezone.utc,
                )

        if value.tzinfo is None:
            value = value.replace(tzinfo=datetime.timezone.utc)
        return long(value.timestamp() * 1000000)

    def convert_back(self, writers_schema, readers_schema, value):
        value = long(value) / 1000000.0
        utc = datetime.datetime.fromtimestamp(value, tz=datetime.timezone.utc)
        return utc

    def does_match(self, writers_schema, readers_schema):
        if isinstance(writers_schema, schema.PrimitiveSchema):
            if (
                writers_schema.type
                in TimestampMicrosLogicalTypeProcessor._matching_types
            ):
                return True
        return False

    def typename(self):
        return "datetime.datetime"

    def initializer(self, value=None):
        return (
            (
                "logical.TimestampMicrosLogicalTypeProcessor().convert_back(None, None, %s)"
                % value
            )
            if value is not None
            else "datetime.datetime.now()"
        )


class TimestampMillisLogicalTypeProcessor(TimestampMicrosLogicalTypeProcessor):
    def convert(self, writers_schema, value):
        return (
            super(TimestampMillisLogicalTypeProcessor, self).convert(
                writers_schema, value
            )
            // 1000
        )

    def convert_back(self, writers_schema, readers_schema, value):
        return super(TimestampMillisLogicalTypeProcessor, self).convert_back(
            writers_schema, readers_schema, value * 1000
        )

    def initializer(self, value=None):
        return (
            (
                "logical.TimestampMillisLogicalTypeProcessor().convert_back(None, None, %s)"
                % value
            )
            if value is not None
            else "datetime.datetime.now()"
        )


DEFAULT_LOGICAL_TYPES = {
    "decimal": DecimalLogicalTypeProcessor(),
    "date": DateLogicalTypeProcessor(),
    "time-millis": TimeMillisLogicalTypeProcessor(),
    "time-micros": TimeMicrosLogicalTypeProcessor(),
    "timestamp-millis": TimestampMillisLogicalTypeProcessor(),
    "timestamp-micros": TimestampMicrosLogicalTypeProcessor(),
}


class LogicalDatumReader(io.DatumReader):
    def __init__(
        self,
        writers_schema=None,
        readers_schema=None,
        logical_types=DEFAULT_LOGICAL_TYPES,
    ):
        """
        Initializes DatumReader with logical type support

        :param schema.Schema writers_schema: Writer's schema
        :param schema.Schema readers_schema: Optional reader's schema
        :param dict[str, LogicalTypeProcessor] logical_types: Optional logical types dict
        """
        super(LogicalDatumReader, self).__init__(
            writers_schema=writers_schema, readers_schema=readers_schema
        )
        self.logical_types = logical_types or {}

    def read_data(self, writers_schema, readers_schema, decoder):
        """
        Reads data from the decoded stream and applies logical type conversion

        :param schema.Schema writers_schema:
        :param schema.Schema readers_schema:
        :param io.BinaryDecoder decoder:
        :return:
        """
        result = super(LogicalDatumReader, self).read_data(
            writers_schema, readers_schema, decoder
        )
        logical_type = readers_schema.props.get("logicalType")
        if logical_type:
            logical_type_handler = self.logical_types.get(logical_type)
            if logical_type_handler and logical_type_handler.does_match(
                writers_schema, readers_schema
            ):
                result = logical_type_handler.convert_back(
                    writers_schema, readers_schema, result
                )
        return result


class LogicalDatumWriter(io.DatumWriter):
    """
    Initializes DatumWriter with logical type support

    :param schema.Schema writers_schema: Writer's schema
    :param dict[str, LogicalTypeProcessor] logical_types: Optional logical types dict
    """

    def __init__(self, writers_schema=None, logical_types=DEFAULT_LOGICAL_TYPES):
        super(LogicalDatumWriter, self).__init__(writers_schema=writers_schema)
        self.logical_types = logical_types

    def write_data(self, writers_schema, datum, encoder):
        logical_type = writers_schema.props.get("logicalType")
        if logical_type:
            logical_type_handler = self.logical_types.get(logical_type)
            if logical_type_handler and logical_type_handler.can_convert(
                writers_schema
            ):
                return super(LogicalDatumWriter, self).write_data(
                    writers_schema,
                    logical_type_handler.convert(writers_schema, datum),
                    encoder,
                )
        return super(LogicalDatumWriter, self).write_data(
            writers_schema, datum, encoder
        )

    def __validate(self, writers_schema, datum):
        logical_type = writers_schema.props.get("logicalType")
        if logical_type:
            lt = self.logical_types.get(logical_type)
            if lt:
                if lt.can_convert(writers_schema):
                    if lt.validate(writers_schema, datum):
                        return True
                    return False

        schema_type = writers_schema.type
        if schema_type == "array":
            return isinstance(datum, list) and False not in [
                self.__validate(writers_schema.items, d) for d in datum
            ]
        elif schema_type == "map":
            return (
                isinstance(datum, dict)
                and False not in [isinstance(k, basestring) for k in datum.keys()]
                and False
                not in [
                    self.__validate(writers_schema.values, v) for v in datum.values()
                ]
            )
        elif schema_type in ["union", "error_union"]:
            return True in [self.__validate(s, datum) for s in writers_schema.schemas]
        elif schema_type in ["record", "error", "request"]:
            return isinstance(datum, dict) and False not in [
                self.__validate(f.type, datum.get(f.name))
                for f in writers_schema.fields
            ]

        return io.validate(writers_schema, datum)

    def write(self, datum, encoder):
        # validate datum
        if not self.__validate(self.writers_schema, datum):
            raise io.AvroTypeException(self.writers_schema, datum)

        self.write_data(self.writers_schema, datum, encoder)


def patch_logical_types():
    io.DatumWriter = LogicalDatumWriter
    io.DatumReader = LogicalDatumReader



import json
import os
import six
from avro import schema

from io import StringIO

from avro.schema import make_avsc_object

from .core_writer import generate_namespace_modules
from .core_writer import clean_fullname
from .tabbed_writer import TabbedWriter
from .core_writer import write_preamble
from .core_writer import write_schema_record
from .core_writer import write_enum
from .core_writer import write_read_file
from .core_writer import write_get_schema
from .core_writer import write_reader_impl
import logging

logger = logging.getLogger("avrogen.schema")
logger.setLevel(logging.INFO)


def generate_schema(
    schema_json, use_logical_types=False, custom_imports=None, avro_json_converter=None
):
    """
    Generate file containing concrete classes for RecordSchemas in given avro schema json
    :param str schema_json: JSON representing avro schema
    :param list[str] custom_imports: Add additional import modules
    :param str avro_json_converter: AvroJsonConverter type to use for default values
    :return Dict[str, str]:
    """

    if avro_json_converter is None:
        avro_json_converter = "avrojson.AvroJsonConverter"

    if "(" not in avro_json_converter:
        avro_json_converter += (
            f"(use_logical_types={use_logical_types}, schema_types=__SCHEMA_TYPES)"
        )

    custom_imports = custom_imports or []
    names = schema.Names()
    make_avsc_object(json.loads(schema_json), names)

    names = [
        k
        for k in six.iteritems(names.names)
        if isinstance(k[1], (schema.RecordSchema, schema.EnumSchema))
    ]
    names = sorted(names, key=lambda x: x[0])

    main_out = StringIO()
    writer = TabbedWriter(main_out)

    write_preamble(writer, use_logical_types, custom_imports)
    write_schema_preamble(writer)
    write_get_schema(writer)
    write_populate_schemas(writer)

    current_namespace = tuple()

    for name, field_schema in names:  # type: str, schema.Schema
        name = clean_fullname(name)
        namespace = tuple(name.split(".")[:-1])
        if namespace != current_namespace:
            current_namespace = namespace
        if isinstance(field_schema, schema.RecordSchema):
            logger.debug(f"Writing schema: {clean_fullname(field_schema.fullname)}")
            write_schema_record(field_schema, writer, use_logical_types)
        elif isinstance(field_schema, schema.EnumSchema):
            logger.debug(
                f"Writing enum: {field_schema.fullname}", field_schema.fullname
            )
            write_enum(field_schema, writer)
    writer.set_tab(0)
    writer.write("\n__SCHEMA_TYPES = {")
    writer.tab()

    # Lookup table for fullname.
    for name, field_schema in names:
        n = clean_fullname(field_schema.name)
        full = field_schema.fullname
        writer.write(f"\n'{full}': {n}Class,")

    # Lookup table for names without namespace.
    for name, field_schema in names:
        n = clean_fullname(field_schema.name)
        writer.write(f"\n'{n}': {n}Class,")

    writer.untab()
    writer.write("\n}\n\n")

    writer.write(f"_json_converter = {avro_json_converter}\n")
    writer.write("avrojson.set_global_json_converter(_json_converter)\n\n")

    value = main_out.getvalue()
    main_out.close()
    return value, [clean_fullname(name[0]) for name in names]


def write_schema_preamble(writer):
    """
    Writes a schema-specific preamble: __get_names_and_schema() which is used by concrete classes to resolve
    their own RecordSchema
    :param writer:
    :return:
    """
    write_read_file(writer)
    writer.write("\n\ndef __get_names_and_schema(json_str):")
    with writer.indent():
        writer.write("\nnames = avro_schema.Names()")
        writer.write("\nschema = make_avsc_object(json.loads(json_str), names)")
        writer.write("\nreturn names, schema")
    writer.write(
        '\n\n\n_SCHEMA_JSON_STR = __read_file(os.path.join(os.path.dirname(__file__), "schema.avsc"))'
    )
    writer.write("\n\n\n__NAMES, _SCHEMA = __get_names_and_schema(_SCHEMA_JSON_STR)")


def write_populate_schemas(writer):
    """
    Writes out a __SCHEMAS dict which contains all RecordSchemas by their full name. Used by get_schema_type
    :param writer:
    :return:
    """
    writer.write(
        '\n__SCHEMAS = dict((n.fullname.lstrip("."), n) for n in six.itervalues(__NAMES.names))\n'
    )


def write_namespace_modules(ns_dict, output_folder):
    """
    Writes content of the generated namespace modules. A python module will be created for each namespace
    and will import concrete schema classes from SchemaClasses
    :param ns_dict:
    :param output_folder:
    :return:
    """
    for ns in six.iterkeys(ns_dict):
        with open(
            os.path.join(output_folder, ns.replace(".", os.path.sep), "__init__.py"),
            "w+",
        ) as f:
            currency = "."
            if ns != "":
                currency += "." * len(ns.split("."))
            for name in ns_dict[ns]:
                f.write(f"from {currency}schema_classes import {name}Class\n")

            f.write("\n\n")

            for name in ns_dict[ns]:
                f.write(f"{name} = {name}Class\n")


def write_specific_reader(record_types, output_folder, use_logical_types):
    """
    Writes specific reader for a avro schema into generated root module
    :param record_types:
    :param output_folder:
    :return:
    """
    with open(os.path.join(output_folder, "__init__.py"), "a+") as f:
        writer = TabbedWriter(f)
        writer.write("from typing import cast")
        writer.write("\nfrom avrogen.dict_wrapper import DictWrapper")
        writer.write("\nfrom .schema_classes import _SCHEMA as get_schema_type")
        writer.write("\nfrom .schema_classes import _json_converter as json_converter")
        for t in record_types:
            writer.write(f'\nfrom .schema_classes import {t.split(".")[-1]}Class')
        writer.write("\nfrom avro.io import DatumReader")
        if use_logical_types:
            writer.write("\nfrom avrogen import logical")

        write_reader_impl(record_types, writer, use_logical_types)


def write_schema_files(
    schema_json, output_folder, use_logical_types=False, custom_imports=None
):
    """
    Generates concrete classes, namespace modules, and a SpecificRecordReader for a given avro schema
    :param str schema_json: JSON containing avro schema
    :param str output_folder: Folder in which to create generated files
    :param list[str] custom_imports: Add additional import modules
    :return:
    """
    schema_py, names = generate_schema(schema_json, use_logical_types, custom_imports)
    names = sorted(names)

    if not os.path.isdir(output_folder):
        os.mkdir(output_folder)

    with open(os.path.join(output_folder, "schema_classes.py"), "w+") as f:
        f.write(schema_py)

    with open(os.path.join(output_folder, "schema.avsc"), "w+") as f:
        f.write(schema_json)

    ns_dict = generate_namespace_modules(names, output_folder)

    with open(os.path.join(output_folder, "__init__.py"), "w+") as f:
        pass  # make sure we create this file from scratch

    write_namespace_modules(ns_dict, output_folder)
    write_specific_reader(names, output_folder, use_logical_types)



from .schema import generate_schema
from .schema import write_schema_files
from .protocol import generate_protocol
from .protocol import write_protocol_files

__all__ = [
    "generate_schema",
    "generate_protocol",
    "write_schema_files",
    "write_protocol_files",
]
